#include <sys/stat.h>
#include <fstream>
#include <iomanip>
#include <string>
#include <iostream>
#include <octave/oct.h>
#include <octave/dColVector.h>
#include <octave/dMatrix.h>
#include <octave/EIG.h>
#include "channelflow/flowfield.h"
#include "channelflow/vector.h"
#include "channelflow/chebyshev.h"
#include "channelflow/tausolver.h"
#include "channelflow/dns.h"
#include "channelflow/symmetry.h"

#include "channelflow/quicksort.h"
#include "octutils.h"

using namespace std;
using namespace channelflow;


void sf(const FlowField& u, Real T, const FieldSymmetry& sigma, FlowField& sfu, 
	Real Reynolds, DNSFlags& flags, TimeStep& dt, int& fcount);

void Dsf(const FlowField& u, const FlowField& sfu, const FlowField& du, FlowField& Df_du,
	 Real T, const FieldSymmetry& sigma, Real Reynolds, DNSFlags& flags,
	 TimeStep& dt, Real eps, bool centerdiff, int& fcount);

void checkConjugacy(const ComplexColumnVector& u, const ComplexColumnVector& v);

int main(int argc, char* argv[]) {

  string purpose("computes eigenvalues of equilibrium or orbit with Arnoldi iteration");
  ArgList args(argc, argv, purpose);

  const bool poincare = args.getflag("-poinc",  "--poincare",    "computing eigenvalues of map on I-D=0 Poincare section");
  //const int  Neig     = args.getint ("-Ne", "--Neigval", 20, "number of eigenvalues/functions to save");
  //const int  n          = args.getint ("-n",  "--neigval", 1,  "check this eigenvalue");

  const Real T        = args.getreal("-T",  "--maptime", 10.0, "integration time for Navier-Stokes.");
  const string sigstr = args.getstr ("-sigma", "--sigma", "", "file containing sigma of sigma f^T(u) - u = 0 (default == identity)");
  const string symmstr = args.getstr ("-symms", "--symmetries", "", "file containing generators of isotropy group for symmetry-constrained search");
  const Real Reynolds = args.getreal("-R", "--Reynolds", 400, "Reynolds number");

  const string nonlstr= args.getstr ("-nl", "--nonl", "rotational", "method for u dot grad u calculation.");
  const string stepstr= args.getstr ("-ts", "--timestepping", "sbdf3", "timestepping algorithm.");


  const bool vardt     = args.getflag("-vdt","--variableDt", "adjust dt to keep CFLmin<=CFL<CFLmax");
  const Real dtarg     = args.getreal("-dt","--timestep",  0.03125, "(initial) integration timestep");
  const Real dtmin     = args.getreal("-dtmin", "--dtmin", 0.001, "minimum timestep");
  const Real dtmax     = args.getreal("-dtmax", "--dtmax", dtarg,  "maximum timestep");
  const Real dTCFL     = args.getreal("-dTCFL", "--dTCFL", 1.00,  "check CFL # at this interval");
  const Real CFLmin    = args.getreal("-CFLmin", "--CFLmin", 0.50, "minimum CFL number");
  const Real CFLmax    = args.getreal("-CFLmax", "--CFLmax", 0.70, "maximum CFL number");

  //const string idir    = args.getpath("-i", "--inputdir", "./", "directory containing eigenfunction data");
  const Real eps      = args.getreal("-eps","--epsilon", 1e-7, "magnitude of perturbation along eigenfunction");

  const string idir = args.getpath("-i", "--indir",  "./", "input directory");

  const string uname  = args.getstr(3, "<flowfield file>", "the eqb/tw solution");
  const string ename  = args.getstr(2, "<flowfield file>", "the eigenfunction");
  const string lname  = args.getstr(1, "<eigenvalue file>", "the eigenvalue");


  args.check();
  args.save("./");
  //mkdir(outdir);

  const Real nu = 1.0/Reynolds;
  TimeStep dt(dtarg, dtmin, dtmax, dTCFL, CFLmin, CFLmax, vardt);

  if (!poincare)
    dt.adjust_for_T(T);

  FieldSymmetry sigma_T;       // defaults to identity
  if (sigstr.length() != 0)
    sigma_T = FieldSymmetry(sigstr);

  DNSFlags dnsflags; 
  dnsflags.baseflow     = PlaneCouette;
  dnsflags.timestepping = s2stepmethod(stepstr);
  dnsflags.dealiasing   = DealiasXZ;
  dnsflags.nonlinearity = s2nonlmethod(nonlstr);
  dnsflags.constraint   = PressureGradient;
  dnsflags.dPdx         = 0.0;
  dnsflags.verbosity    = Silent;

  if (symmstr.length() > 0) {
    SymmetryList symms(symmstr);
    cout << "Restricting flow to invariant subspace generated by symmetries" << endl;
    cout << symms << endl;
    dnsflags.symmetries = symms;
  }
  cout << "dnsflags == " << dnsflags << endl;

  int fcount = 0;

  cout << setprecision(17);
  cout << "setting up DNS and initial fields..." << endl;
  FlowField ustar(uname);

  Complex lambda;
  load(lambda, lname);
  Real mu = Re(lambda);
  Real omega = Im(lambda);

  bool real_eigval = (Im(lambda) == 0.0) ? true : false;

  ofstream os("growth.asc");
  
  cout << "lambda   == " << lambda << endl;
  cout << "Reynolds == " << Reynolds << endl;
  cout << "sigma    == " << sigma_T << endl;
  cout << "T        == " << T << endl;

  if (real_eigval) {

    FlowField ef(ename);
    ef *= 1/L2Norm(ef);

    // Set udns = ustar + eps ef
    FlowField udns = ef;
    udns *= eps/L2Norm(ef);
    udns += ustar;

    FlowField p(udns.Nx(), udns.Ny(), udns.Nz(), 1, 
		udns.Lx(), udns.Lz(), udns.a(), udns.b());
    DNS dns(udns, nu, dt, dnsflags);

    FlowField du;

    for (Real t=0; t<=T; t += dt.dT()) {
      
      //FieldSymmetry sigma_t = sigma_T;
      //sigma_t *= t/T;

      du = udns;
      //du *= sigma_t;
      du -= ustar;
      
      Real magn_dns = L2IP(du, ef);
      Real magn_lin = eps*exp(mu*t);

      os << magn_dns << " " << magn_lin << endl;
      
      cout << t << " " << magn_dns << " " << magn_lin << endl;

      dns.advance(udns, p, dt.n());
    }
  }
  else {  // complex eigval 
    cout << "write this" << endl;

  }

}
