#ifndef CHFLOW_UTILITIES_H
#define CHFLOW_UTILITIES_H

#include <sys/stat.h>
#include <unistd.h>
#include <fstream>
#include <iomanip>
#include <string>
#include <iostream>
#include <vector>

#include "channelflow/flowfield.h"
#include "channelflow/vector.h"
#include "channelflow/chebyshev.h"
#include "channelflow/tausolver.h"
#include "channelflow/dns.h"
#include "channelflow/symmetry.h"
#include "channelflow/quicksort.h"

namespace channelflow {


void save(const std::string& filebase, Real  T, const FieldSymmetry& tau);
void load(const std::string& filebase, Real& T, FieldSymmetry& tau);


// Check divergence, dirichlet, normality, and maybe orthogonality of basis
void basischeck(const std::vector<RealProfile>& e, bool orthocheck=false);

// Make directory and set permissions to a+rx
void mkdir(const std::string& dirname);

// Append trailing / if necess
std::string pathfix(const std::string& path);

// Convert time to string in format convenient for filenames.
// Returns 1 or 1.000 or 1.250 as appropriate. Three digits
std::string t2s(Real t, bool decimals);

// Return false, approx, or true.
std::string fuzzyless(Real x, Real eps);

std::string pwd();

std::string clip(const std::string& filename, const std::string& ext);
std::string stub(const std::string& filename, const std::string& ext);

// Produce plots of various slices, etc.
void plotfield(const FlowField& u_, const std::string& outdir,
	       const std::string& label, int xstride=1, int ystride=1,
	       int zstride=1);

void plotspectra(const FlowField& u_, const std::string& outdir,
		 const std::string& label, bool showpadding=false);

// Produce plots of various slices, etc.
void plotxavg(const FlowField& u_, const std::string& outdir,
	      const std::string& label);

// Produce plots of various slices, etc.
void plotyavg(const FlowField& u_, const std::string& outdir,
	      const std::string& label);

// Return a translation that maximizes the s-symmetry of u.
// (i.e. return tau // that minimizes L2Norm(s(tau(u)) - tau(u))).
FieldSymmetry optimizePhase(const FlowField& u, const FieldSymmetry& s,
			    int Nsteps=10, Real residual=1e-13, Real damp=1.0,
			    bool verbose=true, Real x0=0.0, Real z0=0.0);

// Return |P(u)|^2/|u|^2 where P(u) is projection onto
// sign ==  1 : s-symmetric subspace.
// sign == -1 : s-antisymmetric subspace.
Real PuFraction(const FlowField& u, const FieldSymmetry& s, int sign);

// A simple way to get flexible command-line args for chflow programs
// Scales as N^2 with # args, so don't use for hundreds of args.

class ArgList {
public:
  typedef std::string str; // reduces line-length below

  ArgList();
  ArgList(int argc, char* argv[], const str& purpose);

  bool helpmode() const;
  bool errormode() const;
  int remaining() const;  // how many args have yet to be parsed

  // return true if the option is in the args list, false if not
  bool getflag(const str& shortopt, const str& longopt, const str& helpstr);

  bool getbool(const str& shortopt, const str& longopt, const str& helpstr);
  int  getint (const str& shortopt, const str& longopt, const str& helpstr);
  Real getreal(const str& shortopt, const str& longopt, const str& helpstr);
  str  getstr (const str& shortopt, const str& longopt, const str& helpstr);
  str  getpath(const str& shortopt, const str& longopt, const str& helpstr);

  bool getbool(const str& shortopt, const str& longopt, bool defalt,
	       const str& helpstr);
  int  getint (const str& shortopt, const str& longopt, int defalt,
	       const str& helpstr);
  Real getreal(const str& shortopt, const str& longopt, Real defalt,
	       const str& helpstr);
  str  getstr (const str& shortopt, const str& longopt, const str& defalt,
	       const str& helpstr);
  str  getpath(const str& shortopt, const str& longopt, const str& defalt,
	       const str& helpstr);

  // In the following position counts backwards from end of arglist, e.g
  // in "argv[0] arg3 arg2 arg1" the args are numbered as indicated.
  Real getreal(int position, const std::string& name, const str& helpstr);
  str  getstr (int position, const std::string& name, const str& helpstr);

  void save(const str& outdir) const;  // save command-line to file <argv[0]>.args
  void check();  // check for unrecognized options and arguments

private:
  array<str> args_;
  array<bool> used_;
  bool helpmode_;
  bool errormode_;
  void printhelp(const str& sopt, const str& lopt, const str& type,
		 const str& defalt, const str& helpstr);
  void printhelp(int position, const str& name, const str& helpstr);
};

void field2vector(const FlowField& u, Vector& v);

void vector2field(const Vector& v, FlowField& u);



} // namespace channelflow

#endif // UTILITIES_H
